#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "ppport.h"

typedef struct {
    U32 idx_i;
    U32 idx_j;
    U8 sbox[256];
} RC4_CTX;

int Setup(pTHX_ SV* key, RC4_CTX* ctx)
{
    int i, j, t;
    unsigned char karr[256];
    STRLEN key_len;
    unsigned char *skey = SvPVbyte(key, key_len);

    if (key_len > 0) {
	ctx->idx_i = ctx->idx_j = 0;

	for (i=0; i<256; i++) {
	    ctx->sbox[i] = i;
	}
	for (i=0; i<256; i++) {
	    karr[i] = skey[i%key_len];
	}
	for (i=j=0; i<256; i++) {
	    j = (j + ctx->sbox[i] + karr[i]) % 256;
	    t = ctx->sbox[i];
	    ctx->sbox[i] = ctx->sbox[j];
	    ctx->sbox[j] = t;
	}
	return(1);
    }
    return(0);
}


MODULE = MyRC4		PACKAGE = MyRC4		

void
new(class, key)
    SV* class
    SV*	key

    PREINIT:
	RC4_CTX* ctx;
	STRLEN len;
	int ret;
    PPCODE:
	char *sclass = SvPV(class, len);
	Newx(ctx, 1, RC4_CTX);
	ret = Setup(aTHX_ key, ctx);
	if (ret) {
	    ST(0) = sv_newmortal();
	    sv_setref_pv(ST(0), sclass, ctx);
	    XSRETURN(1);
        } else {
	    Safefree(ctx);
	    XSRETURN_UNDEF;
	}

void
RC4(self, str)
        SV* self
        SV* str

    PREINIT:
        STRLEN len, slen;
        unsigned char *s = SvPVbyte(str, len);
        unsigned char *buf, *p;
	RC4_CTX* ctx;
	register int i;
	register int j;
	register int t;
	register unsigned char *sbox;
	int ret;
    PPCODE:
	slen = len;
	if (!SvROK(self)) {  /* may be called as function */
	    Newx(ctx, 1, RC4_CTX);
	    ret = Setup(aTHX_ self, ctx);
	    if (!ret) {
		Safefree(ctx);
		XSRETURN_UNDEF;
	    }
	} else {
	    ctx = INT2PTR(RC4_CTX*, SvIV(SvRV(self)));
	}
	i = ctx->idx_i;
	j = ctx->idx_j;
	sbox = ctx->sbox;

	Newx(buf, len, U8);
	p = buf;

	while ( len-- ) {
	    i++;
	    i = i & 255;
	    j += sbox[i];
	    j &= 255;
	    t = sbox[i]; sbox[i] = sbox[j]; sbox[j] = t;
	    *p++ = *s++ ^ sbox[(sbox[i] + sbox[j]) & 255];
	}
	ctx->idx_i = i;
	ctx->idx_j = j;
	ST(0) = sv_2mortal(newSVpvn(buf, slen));
	Safefree(buf);
	XSRETURN(1);

void
DESTROY(self)
	SV* self
    CODE:
	if (SvROK(self)) {
	    RC4_CTX* ctx = INT2PTR(RC4_CTX*, SvIV(SvRV(self)));
	    Safefree(ctx);
	}
